#include "include.h"
#include "log.h"
#include "data.h"
#include "socket.h"
#include "socket_msg.h"
#include "socket_handler.h"

static struct uloop_fd server;
int info_sock = 0;
struct sockaddr_in6 own_addr;
struct sockaddr_in6 multicast_addr;

int socket_info(struct socket_msg *info) {
	int ret = 0;
	char msg[REQUEST_MAXLEN];

	ret = socket_msg_marshal(info, msg);
	if (ret) {
		log_error("socket_info: could not marshel message: %d\n", ret);
		return ret;
	}

	ret = sendto(info_sock, msg, REQUEST_MAXLEN, 0, (struct sockaddr *)&multicast_addr, sizeof(multicast_addr));
	if (ret < 0) {
		log_error("socket_info: could not send message: %d\n", ret);
	}
	if(info->type & SOCKET_MSG_TYPE_CLIENT)
		log_verbose("net-send(type=client mac="MACSTR" freq=%d ssi_l=%d ssi_h=%d tprobe=%d tauth=%d authed=%s time=%d)\n", MAC2STR(info->client->addr), info->client->freq_highest, info->client->signal_lowfreq, info->client->signal_highfreq, info->client->try_probe, info->client->try_auth, BOOL2STR(info->client->authed), info->client->time);
	return ret;
}

int socket_request_timeout(struct socket_msg *request, struct socket_msg *answer, struct timeval timeout) {
	int ret = 0;
	int sock = 0;
	char msg[REQUEST_MAXLEN];
	request->type = (request->type | SOCKET_MSG_TYPE_REQUEST);

	ret = socket_msg_marshal(request, msg);
	if (ret) {
		log_error("socket_request: could not marshel message: %d\n", ret);
		return ret;
	}

	sock = socket(AF_INET6, SOCK_DGRAM, IPPROTO_UDP);
	if (sock < 0) {
		log_error("socket_request: could not create socket\n");
		return -1;
	}

	if(bind(sock, (struct sockaddr*)&own_addr, sizeof(own_addr))) {
		log_error("socket_request: could not bind to addr\n");
		return -1;
	}


	int loop = 0;
	if(setsockopt(sock, IPPROTO_IPV6, IPV6_MULTICAST_LOOP, &loop, sizeof(loop)) < 0) {
		log_error("socket_request: unable to disable listen/loop own multicast package: %d\n", errno);
		return -1;
	}

	ret = sendto(sock, msg, REQUEST_MAXLEN, 0, (struct sockaddr *)&multicast_addr, sizeof(multicast_addr));
	if (ret < 0) {
		close(sock);
		log_error("socket_request: could not send message: %d\n", ret);
		return ret;
	}
	if(request->type & SOCKET_MSG_TYPE_CLIENT)
		log_verbose("net-send(type=request,client mac="MACSTR" freq=%d ssi_l=%d ssi_h=%d tprobe=%d tauth=%d authed=%s time=%d)\n", MAC2STR(request->client->addr), request->client->freq_highest, request->client->signal_lowfreq, request->client->signal_highfreq, request->client->try_probe, request->client->try_auth, BOOL2STR(request->client->authed), request->client->time);

	// repeat to recieve best answer till timeout
	ret = setsockopt(sock, SOL_SOCKET, SO_RCVTIMEO, (char*)&timeout,sizeof(timeout));
	if(ret < 0) {
		close(sock);
		log_error("socket_request: unable to set timeout: %d\n", errno);
		return ret;
	}

	ret = recv(sock, msg, REQUEST_MAXLEN, MSG_WAITALL);
	close(sock);
	if (ret < 0) {
		if(errno == EWOULDBLOCK) {
			log_verbose("socket_request: could not recv message - timeout: %d\n", ret);
			return ret;
		}
		log_error("socket_request: could not recv message: %d\n", ret);
		return ret;
	}

	ret = socket_msg_unmarshal(answer, msg);
	if (ret) {
		log_error("socket_request: could not unmarshel message: %d\n", ret);
		return ret;
	}

	if (answer->type & SOCKET_MSG_TYPE_RESPONSE) {
		if(answer->type & SOCKET_MSG_TYPE_CLIENT)
			log_verbose("net-recv(type=response,client mac="MACSTR" freq=%d ssi_l=%d ssi_h=%d tprobe=%d tauth=%d authed=%s time=%d)\n", MAC2STR(answer->client->addr), answer->client->freq_highest, answer->client->signal_lowfreq, answer->client->signal_highfreq, answer->client->try_probe, answer->client->try_auth, BOOL2STR(answer->client->authed), answer->client->time);
		return 0;
	}
	log_verbose("socket_request: end %d\n", ret);
	return -1;
}

static void server_cb(struct uloop_fd *fd, unsigned int events) {
	int ret;
	struct sockaddr_in6 addr;
	char control[256];
	char buf[REQUEST_MAXLEN];

	struct socket_msg msg = {
		.type = 0,
		.client = malloc(sizeof(struct wifi_client)),
	};


	struct iovec iv = {
		.iov_base = buf,
		.iov_len = sizeof(buf)
	};

	struct msghdr mh = {
		.msg_name = &addr,
		.msg_namelen = sizeof(addr),
		.msg_iov = &iv,
		.msg_iovlen = 1,
		.msg_control = control,
		.msg_controllen = sizeof(control)
	};

	if (recvmsg(server.fd, &mh, MSG_WAITALL) < 0) {
		log_info("socket: invalid package\n");
		return;
	}

	ret = socket_msg_unmarshal(&msg, buf);
	if (ret) {
		log_error("socket: could not unmarshel message: %d\n", ret);
		return;
	}
	log_verbose("net-recv(type=");
	if(msg.type & SOCKET_MSG_TYPE_REQUEST)
		log_verbose("request,");
	if(msg.type & SOCKET_MSG_TYPE_RESPONSE)
		log_verbose("response,");
	if(msg.type & SOCKET_MSG_TYPE_CLIENT)
		log_verbose("client mac="MACSTR" freq=%d ssi_l=%d ssi_h=%d tprobe=%d tauth=%d authed=%s time=%d)\n", MAC2STR(msg.client->addr), msg.client->freq_highest, msg.client->signal_lowfreq, msg.client->signal_highfreq, msg.client->try_probe, msg.client->try_auth, BOOL2STR(msg.client->authed), msg.client->time);

	// response only if message are recieve
	ret = socket_handler(&msg);
	if (ret != SOCKET_HANDLER_RESPONSE) {
		if(ret == SOCKET_HANDLER_ERROR)
			log_error("socket_handler: returned  with %d\n", ret);
		return;
	}

	ret = socket_msg_marshal(&msg, buf);
	if (ret) {
		log_error("socket: could not unmarshel message: %d\n", ret);
		return;
	}

	if (sendto(server.fd, buf, sizeof(buf), 0, (struct sockaddr *) &addr, sizeof(addr)) < 0) {
		log_error("socket: could not send answer\n");
	}
	if(msg.type & SOCKET_MSG_TYPE_CLIENT)
		log_verbose("net-send(type=response,client mac="MACSTR" freq=%d ssi_l=%d ssi_h=%d tprobe=%d tauth=%d authed=%s time=%d)\n", MAC2STR(msg.client->addr), msg.client->freq_highest, msg.client->signal_lowfreq, msg.client->signal_highfreq, msg.client->try_probe, msg.client->try_auth, BOOL2STR(msg.client->authed), msg.client->time);
}

int socket_init(char *ifname, char *socket_bind, char *socket_maddr, char *socket_port) {
	server.cb = server_cb;
	server.fd = usock(USOCK_UDP | USOCK_SERVER | USOCK_IPV6ONLY | USOCK_NUMERIC, "::", socket_port);
	if (server.fd < 0) {
		log_error("socket: error create socket to recv\n");
		return 1;
	}
	info_sock = usock(USOCK_UDP | USOCK_IPV6ONLY | USOCK_NUMERIC, socket_bind, socket_port);
	if (info_sock < 0) {
		log_error("socket: error create socket to send\n");
		return 1;
	}

	multicast_addr.sin6_family = AF_INET6;
	multicast_addr.sin6_port = htons(atoi(socket_port));
	multicast_addr.sin6_scope_id = if_nametoindex(ifname);

	own_addr.sin6_family = AF_INET6;
	own_addr.sin6_scope_id = multicast_addr.sin6_scope_id;

	if(!inet_pton(AF_INET6, socket_bind, &own_addr.sin6_addr)) {
		log_error("socket: could not parse address for socket_request\n");
		return -1;
	}

	// listen multicast
	if (!inet_pton(AF_INET6, socket_maddr, &multicast_addr.sin6_addr)) {
		log_error("socket: invalid multicast group\n");
		return -1;
	}

	if (socket_join_mcast(ifname) < 0) {
		log_error("socket: unable to join multicast group: %d\n", errno);
		return -1;
	}

	int loop = 0;
	if(setsockopt(info_sock, IPPROTO_IPV6, IPV6_MULTICAST_LOOP, &loop, sizeof(loop)) < 0) {
		log_error("socket_request: unable to disable listen/loop own multicast package: %d\n", errno);
		return -1;
	}
	if(setsockopt(server.fd, IPPROTO_IPV6, IPV6_MULTICAST_LOOP, &loop, sizeof(loop)) < 0) {
		log_error("socket_request: unable to disable listen/loop own multicast package: %d\n", errno);
		return -1;
	}
	// end listen multicast
	uloop_fd_add(&server, ULOOP_READ);

	return 0;
}

int socket_join_mcast(char *ifname) {
	struct ipv6_mreq mreq = {
		.ipv6mr_multiaddr = multicast_addr.sin6_addr,
		.ipv6mr_interface = if_nametoindex(ifname),
	};
	return setsockopt(server.fd, IPPROTO_IPV6, IPV6_JOIN_GROUP, &mreq, sizeof(mreq));
}

void socket_close() {
	uloop_fd_delete(&server);
	//ustream_free(&(server.fd));
	close(server.fd);
}
